import json
import os

from typing import Optional

import requests

from dataset_sh.constants import DAEMON_PID_FILE
from dataset_sh.server.app import create_app, load_frontend_assets


def save_pid(host: str, port: int, pid: int = None):
    if pid is None:
        pid = os.getpid()

    os.makedirs(os.path.dirname(DAEMON_PID_FILE), exist_ok=True)
    with open(DAEMON_PID_FILE, 'w') as fd:
        daemon_log = {
            'pid': pid,
            'host': host,
            'port': port,
        }
        print(daemon_log)
        json.dump(daemon_log, fd)


def get_running_server_port() -> Optional[tuple[str, int]]:
    if os.path.exists(DAEMON_PID_FILE):
        with open(DAEMON_PID_FILE) as fd:
            try:
                data = json.load(fd)
                host = data['host']
                port = data['port']
                return host, port
            except json.decoder.JSONDecodeError:
                pass
    return None


def daemon_is_running():
    host_port = get_running_server_port()
    if host_port:
        host, port = host_port
        try:
            resp = requests.get(f'http://{host}:{port}/api/version')
            resp.raise_for_status()
            resp_body = resp.json()
            if 'version' in resp_body:
                if resp_body['server'] == 'dataset.sh server':
                    return host, port
        except requests.exceptions.ConnectionError:
            pass

    return None


def launch_daemon_app(host, port) -> Optional[int]:
    _frontend_assets = load_frontend_assets()
    app = create_app(frontend_assets=_frontend_assets)
    save_pid(host, port)
    app.run(port=port, host=host)



import click

from dataset_sh.clients.locator import DatasetLocator
from dataset_sh.clients.obj import Remote, RemoteDataset, LocalStorage, remote
from dataset_sh.utils.misc import parse_tags


class RemoteCtx(object):
    remote: Remote

    def __init__(self, host, profile):
        self.remote = remote(host, profile)


@click.group(name='remote')
@click.option('--host', '-h', 'host', help='host address or alias.', default=None)
@click.option('--profile', '-p', 'profile_name', help='profile name.', default=None)
@click.pass_context
def remote_cli(ctx, host, profile_name):
    """to interact with dataset.sh remote server"""
    ctx.obj = RemoteCtx(host, profile_name)


@remote_cli.command(name='set-tag')
@click.argument('name')
@click.argument('tag')
@click.argument('version')
@click.pass_obj
def tag_remote(remote_ctx: RemoteCtx, name, version, tag):
    """set a version tag for a dataset on a remote server."""
    remote_ctx.remote.dataset(name).set_tag(tag, version)


@remote_cli.command(name='tag-info')
@click.argument('name')
@click.argument('tag')
@click.pass_obj
def print_tag_remote(remote_ctx: RemoteCtx, name, tag):
    """resolve a dataset version tag to version"""
    result = remote_ctx.remote.dataset(name).resolve_tag(tag)
    click.echo(result.version)


@remote_cli.command(name='list-versions')
@click.argument('name')
@click.pass_obj
def list_remote_versions(remote_ctx: RemoteCtx, name):
    """List all available versions of a dataset on a remote server"""

    versions = remote_ctx.remote.dataset(name).versions()
    for v in versions.items:
        click.echo(v.version)


@remote_cli.command(name='download')
@click.argument('name')
@click.option('--file', '-f', default=None,
              help='if this option is provided, the dataset will be downloaded to a file.')
@click.option('--dest', '-d', default=None,
              help='download to local and rename to this name. if file option is provided, this will be ignored.')
@click.option('--tag', '-t', help='tag this version in local.', default='')
@click.pass_obj
def download_remote(remote_ctx: RemoteCtx, name, dest, file, tag):
    """
        download dataset [NAME] from remote.
    """
    remote_dataset_version = DatasetLocator.from_str(name).resolve_remote(remote_ctx.remote)
    if file:
        remote_dataset_version.download_to_file(file)
    else:
        if dest is None:
            dest = name
        local_dataset = DatasetLocator.from_str(dest).resolve_local_dataset()
        remote_dataset_version.download_to(local_dataset, parse_tags(tag))


@remote_cli.command(name='upload')
@click.argument('target_name')
@click.option('--source', '-s', help='use this option to select version to upload', default=None)
@click.option('--tag', '-t', help='tag name of this version in remote.', default='')
@click.pass_obj
def upload(remote_ctx: RemoteCtx, target_name, source, tag):
    """upload managed dataset to server"""
    remote_dataset = DatasetLocator.from_str(target_name).resolve_remote_dataset(remote_ctx.remote)
    if source is None:
        source = target_name
    local_source = DatasetLocator.from_str(source).resolve_local()
    remote_dataset.upload_from_local(local_source, parse_tags(tag))
    local_dataset_source = DatasetLocator.from_str(source).resolve_local_dataset()
    readme = local_dataset_source.readme()
    if readme:
        remote_dataset.update_readme(readme)
    # remote_dataset.upload_from_local(local_source, parse_tags(tag))


@remote_cli.command(name='upload-file')
@click.argument('file')
@click.argument('target_name')
@click.option('--tag', '-t', help='tag name of this version in remote.', default='')
@click.pass_obj
def upload_file(remote_ctx: RemoteCtx, file, target_name, tag):
    """upload dataset to server from a dataset file"""
    remote_dataset = DatasetLocator.from_str(target_name).resolve_remote_dataset(remote_ctx.remote)
    remote_dataset.upload_from_file(file, parse_tags(tag))


@remote_cli.command(name='test-connection')
@click.pass_obj
def test_connection(remote_ctx: RemoteCtx):
    """test connection to remote"""
    user_info = remote_ctx.remote.test_connection()
    username = user_info['username']
    if username:
        click.echo(f"Logged in as {username}")
    else:
        click.echo(f"You are a visitor to this server")



import zipfile


def repackage_zip(input_zip_path, output_zip_path):
    # Open the original zip file in read mode
    with zipfile.ZipFile(input_zip_path, 'r') as input_zip:
        # Create a new zip file with the desired compression (deflate)
        with zipfile.ZipFile(output_zip_path, 'w', compression=zipfile.ZIP_DEFLATED) as output_zip:
            # Iterate through each file in the original zip
            for zip_info in input_zip.infolist():
                # Extract the file's content from the original zip
                with input_zip.open(zip_info.filename) as file:
                    file_data = file.read()

                # Write the content to the new zip file using deflate compression
                output_zip.writestr(zip_info, file_data)

    print(f"Repackaging complete: {output_zip_path}")



import json
import os
import warnings

from dataset_sh.server.core import RepoServerConfig


def load_config(app=None) -> RepoServerConfig:
    config_file = os.environ.get('DATASET_SH_SERVER_CONFIG_FILE', './dataset-sh-server-config.json')

    if config_file is None or not os.path.exists(config_file):
        if not os.path.exists(config_file):
            warnings.warn('server config do not exist, server will be using a default configuration file, '
                          'which may affect some functionality, consider create and modify a configuration file.')
        # host = 'localhost'
        # port = '5000'
        # if app is not None:
        #     host = app.config.get('SERVER_NAME', host)
        #     if host is None:
        #         host = '127.0.0.1'
        #     port = app.config.get('SERVER_PORT', port)
        return RepoServerConfig().override_from_env()
    else:
        with open(config_file) as fd:
            return RepoServerConfig(**json.load(fd)).override_from_env()



from datetime import date, datetime, time, timedelta
from enum import IntEnum, Enum
from typing import get_args, get_origin, get_type_hints, Optional

import typing
from pydantic import BaseModel

from .core import EnumEntry, FieldType, PydanticFieldDefinition, EnumDefinition, ClassDefinition, \
    DatasetSchema


def strip_doc(doc_or_none: Optional[str], from_enum=False) -> Optional[str]:
    if doc_or_none is None:
        return None
    else:
        lines = doc_or_none.strip().split('\n')
        doc_str = '\n'.join([l.strip() for l in lines])
        if from_enum and doc_str == 'An enumeration.':
            return None
        return doc_str


class SchemaBuilder:
    def __init__(self):
        self.classes = []
        self.entry_point = ''
        self.visited = set()

    @staticmethod
    def build(model_clz) -> DatasetSchema:
        builder = SchemaBuilder()
        builder.load_main_model(model_clz)
        class_docstring = model_clz.__doc__
        return DatasetSchema(
            entry_point=builder.entry_point,
            classes=builder.classes,
            doc_str=class_docstring
        )

    def load_main_model(self, clz):
        self.entry_point = get_class_full_name(clz)
        self.load_model(clz)

    def load_model(self, clz):
        """
        load a pydantic model definition. (i.e. any class that inherits BaseModel)
        :param clz:
        :return:
        """
        full_name = get_class_full_name(clz)

        if full_name in self.visited:
            return

        self.visited.add(full_name)

        fields = []
        for field_name, field_type in get_type_hints(clz).items():
            t = self.parse_and_load_definition(field_type)
            fields.append(
                PydanticFieldDefinition(name=field_name, type=t)
            )

        c = ClassDefinition(
            name=full_name,
            fields=fields,
            doc_str=strip_doc(clz.__doc__)
        )

        self.classes.append(c)
        return c

    def load_enum(self, clz):
        full_name = get_class_full_name(clz)

        if full_name in self.visited:
            return
        self.visited.add(full_name)

        entries = []
        for name, value in clz.__members__.items():
            entries.append(EnumEntry(name=name, value=value.value))

        c = EnumDefinition(
            name=full_name,
            enum_entries=entries,
            doc_str=strip_doc(clz.__doc__, from_enum=True)
        )
        self.classes.append(c)
        return c

    def parse_and_load_definition(self, clz) -> FieldType:
        if is_pydantic(clz):
            # Load pydantic class
            self.load_model(clz)
            return FieldType.simple(name=get_class_full_name(clz))
        elif is_enum(clz):
            # Load enum def
            self.load_enum(clz)
            return FieldType.simple(name=get_class_full_name(clz))
        else:
            origin = get_origin(clz)
            if origin is not None:
                # Load things like List[xxx], Dict[A,B]
                pt = self.get_parameterized_type(clz, origin)
                if pt is not None:
                    return pt
            else:
                tn = to_primitive_type(clz)
                if tn is not None:
                    return FieldType.simple(name=tn)

        return FieldType.simple(name='typing.Any')

    def get_parameterized_type(self, clz, origin):

        name = None

        if origin is list or origin is typing.List:
            name = 'typing.List'
        elif origin is set or origin is typing.Set:
            name = 'typing.Set'
        elif origin is dict or origin is typing.Dict:
            name = 'typing.Dict'
        elif origin is tuple or origin is typing.Tuple:
            name = 'typing.Tuple'
        elif origin is typing.Union:
            name = 'typing.Union'

        if name is not None:
            args = get_args(clz)
            if name == 'typing.Union':
                if len(args) == 2 and args[-1] is type(None):
                    name = 'typing.Optional'
                    args = [args[0]]

            children = [self.parse_and_load_definition(a) for a in args]
            return FieldType.typed(name=name, children=children)

        # In all other case, we will return none, and let caller handle this.
        return None


def is_pydantic(clz):
    try:
        return issubclass(clz, BaseModel)
    except TypeError:
        return False


def is_enum(clz):
    try:
        return issubclass(clz, Enum) or issubclass(clz, IntEnum)
    except TypeError:
        return False


def to_datetime_types(clz):
    if clz is date:
        return 'datetime.date'
    elif clz is datetime:
        return 'datetime.datetime'
    elif clz is time:
        return 'datetime.time'
    elif clz is timedelta:
        return 'datetime.timedelta'
    return None


def to_primitive_type(clz):
    dt_type = to_datetime_types(clz)
    if dt_type is not None:
        return dt_type

    if clz is int:
        return 'int'
    elif clz is float:
        return 'float'
    elif clz is bool:
        return 'bool'

    elif clz is str:
        return 'str'

    elif clz is list:
        return 'list'

    elif clz is dict:
        return 'dict'

    elif clz is tuple:
        return 'tuple'

    elif clz is set:
        return 'set'

    return None


def get_class_full_name(clz):
    return clz.__module__ + '.' + clz.__name__



import click

from dataset_sh.clients.locator import DatasetLocator
from dataset_sh.clients.obj import Remote, remote
from dataset_sh.dependency.parser import parse_file


@click.group(name='project')
def project_cli():
    """to manage dataset.sh project"""
    pass


@project_cli.command(name='install')
@click.argument('file_path')
def install_project(file_path):
    """install all datasets listed in project file."""

    dep = parse_file(file_path)
    for host_group in dep.dependencies:
        host = host_group.host
        host_remote = remote(host=host)
        for dataset_item in host_group.datasets:
            source_locator = DatasetLocator.from_str(dataset_item.source)

            target = dataset_item.target
            if target is None:
                target = dataset_item.source

            target_locator = DatasetLocator.from_str(target)
            source_dv = source_locator.resolve_remote(host_remote)
            source_dv.download_to(
                target_locator.resolve_local_dataset()
                ,
                tags=[target_locator.tag]
            )



def generate_bash_import(site_url, dataset_full_name):
    url = generate_download_url(site_url, dataset_full_name)
    code = f"""
import dataset_sh
dataset_sh.import_url("{dataset_full_name}", "{url}")   
    """.strip()
    return code


def generate_python_import(site_url, dataset_full_name):
    url = generate_download_url(site_url, dataset_full_name)
    code = f"""
dataset.sh import {dataset_full_name} -u {url}
    """.strip()
    return code


def generate_download_url(site_url, dataset_full_name):
    return f"{site_url}/api/dataset/{dataset_full_name}/file"



from typing import Optional, Union

from dataset_sh.clients.obj import Remote, LocalStorage, LocalDatasetVersion, RemoteDatasetVersion, \
    RemoteDatasetVersionTag, RemoteDataset, LocalDataset


class DatasetLocator:
    namespace: str
    dataset_name: str
    tag: Optional[str] = None
    version: Optional[str] = None

    def __init__(self, namespace: str, dataset_name: str, tag: Optional[str] = None, version: Optional[str] = None):
        self.namespace = namespace
        self.dataset_name = dataset_name
        self.tag = tag
        self.version = version

    @staticmethod
    def from_str(str_value) -> 'DatasetLocator':
        """
        DatasetLocator accept the following locator format:
        1. [namespace]/[dataset_name]
            this will be interpreted as [namespace]/[dataset_name]:tag=latest
        2. [namespace]/[dataset_name]:version=[version]
        3. [namespace]/[dataset_name]:tag=[tag_name]

        Args:
            str_value: The string representation of the dataset locator.

        Returns: An instance of DatasetLocator.

        Raises:
            ValueError: If the provided string is not in a valid format.
        """
        parts = str_value.split(':')
        namespace_dataset = parts[0].split('/')
        if len(namespace_dataset) != 2:
            raise ValueError("Invalid format. Must be in the format [namespace]/[dataset_name]")

        namespace, dataset_name = namespace_dataset

        if len(parts) == 1:
            return DatasetLocator(namespace, dataset_name, tag="latest")
        elif len(parts) == 2:
            option, value = parts[1].split('=')
            if option == 'version':
                return DatasetLocator(namespace, dataset_name, version=value)
            elif option == 'tag':
                return DatasetLocator(namespace, dataset_name, tag=value)
            else:
                raise ValueError("Invalid option. Must be 'version' or 'tag'")
        else:
            raise ValueError("Invalid format")

    def resolve_local(self) -> LocalDatasetVersion:
        dataset = LocalStorage().namespace(self.namespace).dataset(self.dataset_name)
        if self.version:
            dv = dataset.version(self.version)
        elif self.tag:
            dv = dataset.tag(self.tag)
        else:
            dv = dataset.latest()
        return dv

    def resolve_remote(self, remote: Remote) -> Union[RemoteDatasetVersion, RemoteDatasetVersionTag]:
        dataset = remote.namespace(self.namespace).dataset(self.dataset_name)
        if self.version:
            dv = dataset.version(self.version)
        elif self.tag:
            dv = dataset.tag(self.tag)
        else:
            dv = dataset.latest()
        return dv

    def resolve_local_dataset(self) -> LocalDataset:
        dataset = LocalStorage().namespace(self.namespace).dataset(self.dataset_name)
        return dataset

    def resolve_remote_dataset(self, remote: Remote) -> RemoteDataset:
        dataset = remote.namespace(self.namespace).dataset(self.dataset_name)
        return dataset



import binascii
import json
import math
import os
from functools import wraps

import jwt
from flask import Flask, jsonify, send_file, g, request, make_response
from head_switcher import install_to_flask, load_from_package_resources

from dataset_sh.clients import LocalStorage
from dataset_sh.multipart.multipart import MultipartFileWriter
from dataset_sh.utils.files import filesize, file_creation_time
from werkzeug.security import safe_join
from importlib.metadata import version as lib_version

from dataset_sh.server.config import load_config
from dataset_sh.server.project_file_manager import ProjectArticleManager
from dataset_sh.server.utils import generate_download_code
from dataset_sh.utils.usage.read_data_codegen import get_read_data_code
from .core import RepoServerConfig
from .constants import ACCESS_KEY_HEADER_NAME, TOKEN_HEADER_KEY


def load_frontend_assets():
    try:
        return load_from_package_resources('dataset_sh.server.assets', 'app-ui.frontend', prefix='dist/')
    except FileNotFoundError as e:
        return {
            'index.html': 'dataset.sh web interface is disabled.'
        }


DISABLE_UI = os.environ.get('DISABLE_DATASET_APP_UI', '0').lower() in ['true', '1']


def calculate_chunk_size(config, file_length):
    max_chunk_count = config.max_chunk_count
    minimal_chunk_size = config.minimal_chunk_size

    if max_chunk_count * minimal_chunk_size > file_length:
        return minimal_chunk_size
    else:
        chunk_size = int(math.ceil(file_length / max_chunk_count))
        return chunk_size


def create_app(frontend_assets=None, config: RepoServerConfig = None):
    app = Flask(__name__, static_folder=None)

    if config is None:
        config = load_config(app).override_from_env()

    if frontend_assets is None:
        if not DISABLE_UI:
            frontend_assets = load_frontend_assets()
        else:
            frontend_assets = {
                'index.html': "dataset.sh web ui is disabled"
            }

    config.make_dirs()

    local_storage = LocalStorage(config.data_folder)
    pf_manager = ProjectArticleManager(config.article_folder)

    @app.route('/api/version', methods=['GET'])
    def get_server_version():
        return jsonify({
            'server': 'dataset.sh server',
            'version': lib_version('dataset.sh')
        })

    if config.require_auth:
        @app.route('/api/login', methods=['POST'])
        def login():
            username = request.json.get('username')
            password = request.json.get('password')
            if config.verify_userpass(username, password):
                token = jwt.encode({
                    'username': username
                }, config.secret, algorithm="HS256")
                response = make_response(jsonify({'message': 'Login successful'}))
                response.set_cookie(TOKEN_HEADER_KEY, token)
                return response
            else:
                return '', 401

        @app.route('/api/logout', methods=['POST'])
        def logout():
            response = make_response(jsonify({'message': 'Logout successful'}))
            response.delete_cookie(TOKEN_HEADER_KEY)
            return response

    @app.before_request
    def before_request_func():

        g.current_user = None
        token = request.cookies.get(TOKEN_HEADER_KEY, None)
        if token:
            try:
                payload = jwt.decode(token, config.secret, algorithms=['HS256'])
                g.current_user = payload['username']
            except jwt.ExpiredSignatureError:  # pragma: no cover
                pass
            except jwt.InvalidTokenError:  # pragma: no cover
                pass
        else:
            access_key = request.headers.get(ACCESS_KEY_HEADER_NAME, None)
            if access_key is not None:
                try:
                    g.current_user = config.verify_key(access_key)
                except (json.decoder.JSONDecodeError, binascii.Error):
                    pass

    def require_auth(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            if config.require_auth and g.current_user is None:
                return '', 401
            return f(*args, **kwargs)

        return decorated_function

    @app.route('/api/info', methods=['GET'])
    @require_auth
    def get_info():
        if g.current_user is None:
            return jsonify({'username': ''})
        else:
            return jsonify({'username': g.current_user})

    @app.route('/api/hostname', methods=['GET'])
    @require_auth
    def get_host():
        return jsonify({'hostname': config.hostname})

    @app.route('/api/dataset', methods=['GET'])
    @require_auth
    def list_datasets():
        items = local_storage.datasets()
        return jsonify({
            'items': [{
                'namespace': item.namespace,
                'dataset': item.dataset_name,
            } for item in items]
        }), 200

    @app.route('/api/namespace', methods=['GET'])
    @require_auth
    def list_namespaces():
        items = local_storage.namespaces()
        return jsonify({
            'namespaces': [item.namespace for item in items]
        }), 200

    @app.route('/api/dataset/<namespace>', methods=['GET'])
    @require_auth
    def list_datasets_in_store(namespace):
        items = local_storage.namespace(namespace).datasets()
        return jsonify({
            'items': [{
                'namespace': item.namespace,
                'dataset': item.dataset_name,
            } for item in items]
        }), 200

    def _get_version_description(namespace, dataset_name, version):
        ret = {
            "version": version,
        }
        dv = local_storage.namespace(namespace).dataset(dataset_name).version(version)
        if dv.exists():
            meta = dv.meta()

            fp = dv.datafile()
            if os.path.exists(fp):
                ret['fileSize'] = filesize(fp)
                ret['createdAt'] = int(file_creation_time(fp).timestamp())

            d = meta.get('description', None)
            if d:
                ret['description'] = d
        return ret

    @app.route('/api/dataset/<namespace>/<dataset_name>/version', methods=['GET'])
    @require_auth
    def list_dataset_versions(namespace, dataset_name):
        versions = local_storage.namespace(namespace).dataset(dataset_name).versions()
        return jsonify({
            'items': [_get_version_description(namespace, dataset_name, v.version) for v in versions]
        })

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag', methods=['GET'])
    @require_auth
    def list_dataset_tags(namespace, dataset_name):
        tags = local_storage.namespace(namespace).dataset(dataset_name).tags()
        return jsonify({
            'items': [{'tag': tag, 'version': version} for tag, version in tags.items()]
        })

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>', methods=['GET'])
    @require_auth
    def get_tag(namespace, dataset_name, tag):
        t = local_storage.namespace(namespace).dataset(dataset_name).resolve_tag(tag)
        if t:
            return jsonify({
                'tag': tag,
                'version': t
            })
        else:
            return '', 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>', methods=['POST'])
    @require_auth
    def set_tag(namespace, dataset_name, tag):

        if not config.allow_upload:
            return '', 401

        if not hasattr(g, 'current_user') or g.current_user != namespace:
            return '', 401

        local_storage.namespace(namespace).dataset(dataset_name).set_tag(tag, request.json.get('version'))
        return '', 204

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>', methods=['DELETE'])
    @require_auth
    def remove_tag(namespace, dataset_name, tag):
        if not config.allow_upload:
            return '', 401

        if not hasattr(g, 'current_user') or g.current_user != namespace:
            return '', 401
        local_storage.namespace(namespace).dataset(dataset_name).remove_tag(tag)
        return '', 204

    #
    #
    #
    #
    #  By Version
    #
    #
    #
    #
    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>/meta', methods=['GET'])
    @require_auth
    def get_dataset_meta_by_version(namespace, dataset_name, version):
        dv = local_storage.namespace(namespace).dataset(dataset_name).version(version)
        if dv.exists():
            meta = dv.meta()
            fp = dv.datafile()
            meta['fileSize'] = filesize(fp)
            return jsonify(meta), 200
        else:
            return '', 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>', methods=['DELETE'])
    @require_auth
    def r_remove_dataset_version(namespace, dataset_name, version):
        if not config.allow_upload:
            return '', 401
        if not hasattr(g, 'current_user') or g.current_user != namespace:
            return '', 401

        dv = local_storage.namespace(namespace).dataset(dataset_name).version(version)
        if dv.exists():
            dv.delete()
            return '', 204
        else:
            return '', 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>/collection/<collection_name>/sample',
               methods=['GET'])
    @require_auth
    def get_collection_sample_by_version(namespace, dataset_name, version, collection_name):
        sample = local_storage.namespace(namespace).dataset(dataset_name).version(version).sample(collection_name)
        return jsonify(sample), 200

    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>/collection/<collection_name>/type',
               methods=['GET'])
    @require_auth
    def get_collection_ta_by_version(namespace, dataset_name, version, collection_name):
        ta_dict = local_storage.namespace(namespace).dataset(dataset_name).version(version).type_annotation_dict(
            collection_name)
        return jsonify({"type": ta_dict}), 200

    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>/collection/<collection_name>/code',
               methods=['GET'])
    @require_auth
    def get_collection_code_by_version(namespace, dataset_name, version, collection_name):
        reader_code = get_read_data_code(
            f'{namespace}/{dataset_name}',
            collection_name,
            version=version,
            tag=None
        )
        return {'code': reader_code}, 200

    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>/download-code', methods=['GET'])
    @require_auth
    def get_download_code_by_version(namespace, dataset_name, version):
        fullname = f"{namespace}/{dataset_name}"
        lang = request.values.get('lang', 'python')
        if config.hostname:
            code = generate_download_code(lang, config.hostname, fullname, version=version)
            return {'code': code}, 200
        else:
            return {'code': ''}, 200

    @app.route('/api/dataset/<namespace>/<dataset_name>/version/<version>/file', methods=['GET'])
    @require_auth
    def get_dataset_file_by_version(namespace, dataset_name, version):
        dv = local_storage.namespace(namespace).dataset(dataset_name).version(version)
        if dv.exists():
            f = dv.datafile()
            if os.path.exists(f):
                root = os.path.normpath(local_storage.location)
                f = os.path.normpath(f)
                if f.startswith(root):
                    return send_file(
                        f,
                        as_attachment=True,
                        download_name=f"{namespace}_{dataset_name}.dataset"
                    )
        return '', 404

    #
    #
    #
    #
    #  By Tag
    #
    #
    #
    #
    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>/meta', methods=['GET'])
    @require_auth
    def get_dataset_meta_by_tag(namespace, dataset_name, tag):
        dv = local_storage.namespace(namespace).dataset(dataset_name).tag(tag, allow_none=True)
        if dv and dv.exists():
            meta = dv.meta()
            fp = dv.datafile()
            meta['fileSize'] = filesize(fp)
            return jsonify(meta), 200
        else:
            return '', 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>/collection/<collection_name>/sample',
               methods=['GET'])
    @require_auth
    def get_collection_sample_by_tag(namespace, dataset_name, tag, collection_name):
        dv = local_storage.namespace(namespace).dataset(dataset_name).tag(tag, allow_none=True)
        if dv and dv.exists():
            sample = dv.sample(collection_name)
            return jsonify(sample), 200
        else:
            return "", 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>/collection/<collection_name>/type',
               methods=['GET'])
    @require_auth
    def get_collection_typa_anotation_by_tag(namespace, dataset_name, tag, collection_name):
        dv = local_storage.namespace(namespace).dataset(dataset_name).tag(tag, allow_none=True)
        if dv and dv.exists():
            ta_dict = dv.type_annotation_dict(collection_name)
            return jsonify({"type": ta_dict}), 200
        else:
            return "", 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>/collection/<collection_name>/code',
               methods=['GET'])
    @require_auth
    def get_collection_code_by_tag(namespace, dataset_name, tag, collection_name):
        dv = local_storage.namespace(namespace).dataset(dataset_name).tag(tag, allow_none=True)
        if dv and dv.exists():
            reader_code = get_read_data_code(
                f'{namespace}/{dataset_name}',
                collection_name,
                version=None,
                tag=tag
            )
            return jsonify({'code': reader_code}), 200
        else:
            return "", 404

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>/download-code', methods=['GET'])
    @require_auth
    def get_download_code_by_tag(namespace, dataset_name, tag):
        fullname = f"{namespace}/{dataset_name}"
        lang = request.values.get('lang', 'python')
        if config.hostname:
            code = generate_download_code(lang, config.hostname, fullname, tag=tag)
            return {'code': code}, 200
        else:
            return {'code': ''}, 200

    @app.route('/api/dataset/<namespace>/<dataset_name>/tag/<tag>/file', methods=['GET'])
    @require_auth
    def get_dataset_file_by_tag(namespace, dataset_name, tag):
        dv = local_storage.namespace(namespace).dataset(dataset_name).tag(tag, allow_none=True)
        if dv and dv.exists():
            f = dv.datafile()
            if os.path.exists(f):
                root = os.path.normpath(local_storage.location)
                f = os.path.normpath(f)
                if f.startswith(root):
                    return send_file(
                        f,
                        as_attachment=True,
                        download_name=f"{namespace}_{dataset_name}.dataset"
                    )
        return '', 404

    #
    #
    #
    #
    #  Upload Dataset
    #
    #
    #
    #

    @app.route('/api/dataset/<namespace>/<dataset_name>', methods=['POST'])
    @require_auth
    def create_dataset(namespace, dataset_name):
        if config.allow_upload:
            user_name = namespace
            dataset_name = dataset_name
            tags = [t.strip() for t in request.args.get('tag', '').split(',') if t.strip() != '']

            if not hasattr(g, 'current_user') or g.current_user != user_name:
                return '', 401

            checksum_value = request.args['checksum_value']
            file_length = int(request.args['file_length'])
            action = request.args['action']

            filename = f"dataset_file_{checksum_value}"

            os.makedirs(safe_join(config.uploader_folder, user_name, dataset_name), exist_ok=True)
            fp = safe_join(config.uploader_folder, user_name, dataset_name, filename)

            writer = MultipartFileWriter(
                fp,
                checksum=checksum_value,
                file_length=file_length,
                chunk_size=calculate_chunk_size(config, file_length)
            )

            if action == 'init':
                existing = writer.load_existing_progress()
                if existing is None:
                    writer.start()
                    existing = writer.load_existing_progress()
                return jsonify(existing.model_dump(mode='json'))
            elif action == 'upload':
                part_id = request.args['part_id']
                part_file = writer.part_file(int(part_id))
                with open(part_file, 'wb') as out:
                    out.write(request.data)
                writer.mark_file_as_complete(part_id)
                return '', 204
            elif action == 'done':
                writer.done()
                version_hex = local_storage.namespace(namespace).dataset(dataset_name).import_file(
                    fp,
                    remove_source=True,
                    tags=tags
                )
                return '', 204
            elif action == 'abort':
                writer.terminate()
                return '', 204
            return '', 400
        else:
            return '', 401

    @app.route('/api/dataset/<namespace>/<dataset_name>/readme', methods=['POST'])
    @require_auth
    def update_dataset_readme(namespace, dataset_name):
        if config.allow_upload:
            t = local_storage.namespace(namespace).dataset(dataset_name)
            readme = request.json.get('readme')
            t.set_readme(readme)
            return '', 204
        else:
            return '', 401

    @app.route('/api/dataset/<namespace>/<dataset_name>/readme', methods=['GET'])
    @require_auth
    def get_dataset_readme(namespace, dataset_name):
        t = local_storage.namespace(namespace).dataset(dataset_name)
        readme = t.readme()
        if readme is None:
            return '', 404
        return readme, 200

    #
    #
    #
    #
    #  Articles
    #
    #
    #
    #
    @app.route('/api/post-assets/<path:asset_path>', methods=['GET'])
    @require_auth
    def get_asset_path(asset_path):
        fp = os.path.join(
            pf_manager.base_folder,
            '__assets__',
            asset_path
        )

        fp = os.path.normpath(fp)
        if not fp.startswith(pf_manager.base_folder):
            return '', 404

        if os.path.exists(fp):
            return send_file(fp)
        else:
            return '', 404

    @app.route('/api/post', methods=['GET'])
    @require_auth
    def list_posts():
        articles = [a.model_dump(mode='json') for a in pf_manager.list_articles()]
        return jsonify({'items': articles})

    @app.route('/api/post/<path:post_name>', methods=['GET'])
    @require_auth
    def get_post(post_name):
        article = pf_manager.extract_article(post_name)
        if article is None:
            return '', 404
        return article

    install_to_flask(frontend_assets, app)

    return app


if __name__ == '__main__':  # pragma: no cover
    _frontend_assets = {
        'index.html': "dataset.sh web ui is disabled"
    }

    if not DISABLE_UI:
        _frontend_assets = load_frontend_assets()

    app = create_app(frontend_assets=_frontend_assets)
    app.run(port=8989)



import json
import secrets

from werkzeug.security import generate_password_hash, check_password_hash


def generate_password():
    return secrets.token_hex(16)


def encode_key(username, password):
    kid = secrets.token_hex(4)
    key = json.dumps({
        'key_id': kid,
        'username': username,
        'password': password
    })
    return f"dsh-key:{username}:{kid}:{password}"


def decode_key(key_str):
    parts = key_str.split(':')
    if len(parts) != 4:
        return None
    [header, username, key_name, key] = parts
    if header != 'dsh-key':
        return None
    return username, key


def verify_password(password, hashed_password):
    return check_password_hash(hashed_password, password)


def hash_password(password):
    return generate_password_hash(password)
